import pymem, re, time, struct
from __future__ import annotations


class Exploit:
    def __init__(self, ProgramName=None):
        self.ProgramName = ProgramName
        self.Program = pymem.Pymem()
        self.Addresses = {}
        if type(ProgramName) == str:
            self.Program = pymem.Pymem(ProgramName)
        elif type(ProgramName) == int:
            self.Program.open_process_from_id(ProgramName)

    def h2d(self, hz: str) -> int:
        if type(hz) == int:
            return hz
        return int(hz, 16)

    def d2h(self, dc: int) -> str:
        if type(dc) == str:
            return dc
        if abs(dc) > 4294967295:
            dc = hex(dc & (2**64 - 1)).replace("0x", "")
        else:
            dc = hex(dc & (2**32 - 1)).replace("0x", "")
        if len(dc) > 8:
            while len(dc) < 16:
                dc = "0" + dc
        if len(dc) < 8:
            while len(dc) < 8:
                dc = "0" + dc
        return dc

    def PLAT(self, aob: str):
        if type(aob) == bytes:
            return aob
        trueB = bytearray(b"")
        aob = aob.replace(" ", "")
        PLATlist = []
        for i in range(0, len(aob), 2):
            PLATlist.append(aob[i : i + 2])
        for i in PLATlist:
            if "?" in i:
                trueB.extend(b".")
            if "?" not in i:
                trueB.extend(re.escape(bytes.fromhex(i)))
        return bytes(trueB)

    def AOBSCANALL(self, AOB_Sig, xreturn_multiple=False):
        return pymem.pattern.pattern_scan_all(
            self.Program.process_handle,
            self.PLAT(AOB_Sig),
            return_multiple=xreturn_multiple,
        )

    def gethexc(self, hex: str):
        hex = hex.replace(" ", "")
        hxlist = []
        for i in range(0, len(hex), 2):
            hxlist.append(hex[i : i + 2])
        return len(hxlist)

    def hex2le(self, hex: str):
        lehex = hex.replace(" ", "")
        reniL = 0
        zqSij = ""
        lelist = []
        for i in range(0, len(lehex), 2):
            lelist.append(lehex[i : i + 2])
        if len(lelist) != 4:
            reniL = len(lelist) - 4
            zqSij = zqSij + "0"
            for i in range(0, reniL):
                zqSij = zqSij + "00"
        lelist.insert(0, zqSij)
        if len("".join(lelist)) != 8:
            lelist.insert(0, "0")
        lelist.reverse()
        lehex = "".join(lelist)
        return lehex

    def calcjmpop(self, des, cur):
        jmpopc = (self.h2d(des) - self.h2d(cur)) - 5
        jmpopc = hex(jmpopc & (2**32 - 1)).replace("0x", "")
        if len(jmpopc) % 2 != 0:
            jmpopc = "0" + str(jmpopc)
        return jmpopc

    def isProgramGameActive(self):
        try:
            self.Program.read_char(self.Program.base_address)
            return True
        except:
            return False

    def DRP(self, Address: int) -> int:
        Address = Address
        if type(Address) == str:
            Address = self.h2d(Address)
        return int.from_bytes(self.Program.read_bytes(Address, 4), "little")

    def isValidPointer(self, Address: int) -> bool:
        try:
            if type(Address) == str:
                Address = self.h2d(Address)
            self.Program.read_bytes(self.DRP(Address), 1)
            return True
        except:
            return False

    def HookAddressBase(self, AccessLocation, AccessLocationBytes, movOP):
        mem = self.Program
        BaseAddress = 0
        OriginalHexArray = bytes.hex(
            self.Program.read_bytes(AccessLocation, AccessLocationBytes)
        )
        BaseAddressA = mem.allocate(4)
        newmem = mem.allocate(256)
        BAOP = movOP + self.hex2le(self.d2h(BaseAddressA))
        hookLoc = AccessLocation
        returnJmp = "E9" + self.hex2le(
            self.calcjmpop(
                self.d2h(hookLoc + (self.gethexc(OriginalHexArray))),
                self.d2h(newmem + (self.gethexc(BAOP + OriginalHexArray))),
            )
        )
        newmemArrayW = BAOP + OriginalHexArray + returnJmp
        mem.write_bytes(newmem, bytes.fromhex(newmemArrayW), self.gethexc(newmemArrayW))
        mem.write_bytes(
            hookLoc,
            bytes.fromhex(
                "E9 " + self.hex2le(self.calcjmpop(self.d2h(newmem), self.d2h(hookLoc)))
            ),
            AccessLocationBytes,
        )
        BaseAddress = mem.read_int(BaseAddressA)
        HookData = [BaseAddressA, newmem, BaseAddress, OriginalHexArray, AccessLocation]
        return HookData

    def UnHookAddressBase(self, HookData):
        self.Program.write_bytes(
            HookData[4], bytes.fromhex(HookData[3]), self.gethexc(HookData[3])
        )
        self.Program.free(HookData[0])
        self.Program.free(HookData[1])
        return HookData[4]

    def YieldForHookA(self, AAP, amount=0.5, bounds=50):
        looped = 0
        while self.Program.read_int(AAP) == 0:
            if looped > bounds:
                print("Outside maximum loop allowed.")
                return False
            time.sleep(amount)
            looped += 1
        return True

    def GetModules(self) -> list:
        return list(self.Program.list_modules())

    def getAddressFromName(self, Address: str) -> int:
        if type(Address) == int:
            return Address
        AddressBase = 0
        AddressOffset = 0
        for i in self.GetModules():
            if i.name in Address:
                AddressBase = i.lpBaseOfDll
                AddressOffset = self.h2d(Address.replace(i.name + "+", ""))
                AddressNamed = AddressBase + AddressOffset
                return AddressNamed
        print("Unable to find Address: " + Address)
        return Address

    def getNameFromAddress(self, Address: int) -> str:
        memoryInfo = pymem.memory.virtual_query(self.Program.process_handle, Address)
        AllocationBase = memoryInfo.AllocationBase
        NameOfDLL = ""
        AddressOffset = 0
        for i in self.GetModules():
            if i.lpBaseOfDll == AllocationBase:
                NameOfDLL = i.name
                AddressOffset = Address - AllocationBase
            break
        if NameOfDLL == "":
            return Address
        NameOfAddress = NameOfDLL + "+" + self.d2h(AddressOffset)
        return NameOfAddress

    def getRawProcesses(self):
        toreturn = []
        for i in pymem.process.list_processes():
            toreturn.append(
                [
                    i.cntThreads,
                    i.cntUsage,
                    i.dwFlags,
                    i.dwSize,
                    i.pcPriClassBase,
                    i.szExeFile,
                    i.th32DefaultHeapID,
                    i.th32ModuleID,
                    i.th32ParentProcessID,
                    i.th32ProcessID,
                ]
            )
        return toreturn

    def SimpleGetProcesses(self):
        toreturn = []
        for i in self.getRawProcesses():
            toreturn.append({"Name": i[5].decode(), "Threads": i[0], "ProcessId": i[9]})
        return toreturn

    def YieldForProgram(self, programName, AutoOpen: bool = False, Limit=15):
        Count = 0
        while True:
            if Count > Limit:
                print("Yielded too long, failed!")
                return False
            ProcessesList = self.SimpleGetProcesses()
            for i in ProcessesList:
                if i["Name"] == programName:
                    print(
                        "Found "
                        + programName
                        + " with Process ID: "
                        + str(i["ProcessId"])
                    )
                    if AutoOpen:
                        self.Program.open_process_from_id(i["ProcessId"])
                        print("Successfully attached to Process.")
                        return True
            print("Waiting for the Program...")
            time.sleep(1)
            Count += 1

    def GetProcessesByName(self, Name):
        result = []
        for process in self.SimpleGetProcesses():
            if process["Name"] == Name:
                result.append(process)

    def ReadStringUntilEnd(self, Address: int) -> str:
        if type(Address) == str:
            Address = self.h2d(Address)
        if Address == 0:
            return ""
        CurrentAddress = Address
        StringData = []
        LoopedTimes = 0
        while LoopedTimes < 15000:
            if self.Program.read_bytes(CurrentAddress, 1) == b"\x00":
                break
            StringData.append(self.Program.read_bytes(CurrentAddress, 1))
            CurrentAddress += 1
            LoopedTimes += 1
        String = bytes()
        for i in StringData:
            String = String + i
        return str(String)[2:-1]

    def ReadInstaceString(self, Address: int) -> str:
        length = self.Program.read_int(self.DRP(Address) + 0x10)
        if length < 16 and length > 0:
            return self.ReadStringUntilEnd(self.DRP(Address))
        else:
            return self.ReadStringUntilEnd(self.DRP(self.DRP(Address)))

    def ReadNormalString(self, Address: int) -> str:
        length = self.Program.read_int(Address + 0x10)
        if length < 16 and length > 0:
            return self.ReadStringUntilEnd(Address)
        else:
            return self.ReadStringUntilEnd(self.DRP(Address))


def GetProcessIdByNameWithBiggestSize(NameOfProgram: str):
    x = Exploit()
    x.GetProcessesByName(NameOfProgram)
    y = x.SimpleGetProcesses()
    dictionaryOfData = {}
    for i in y:
        if i["Name"] == NameOfProgram:
            x.Program.open_process_from_id(i["ProcessId"])
            Size = x.Program.process_base.SizeOfImage
            dictionaryOfData.update({Size: i["ProcessId"]})
        tempVa = 0
    for i in dictionaryOfData.keys():
        if i > tempVa:
            tempVa = i
    return dictionaryOfData.get(tempVa)


roblox = Exploit(GetProcessIdByNameWithBiggestSize("RobloxPlayerBeta.exe"))

instance_offsets = {
    "name": 0x2C,
    "parent": 0x3C,
    "childlist": 0x30,
    "class_descriptor": 0xC,
    "property_descriptor": 0x18,
    "event_descriptor": 0x78,
    "boundedfunctions": 0xD8,
    "new": "RobloxPlayerBeta.exe+461D70",
    "name_map": "RobloxPlayerBeta.exe+3AE46C0",
}

property_descriptor_offsets = {
    "name": 0x4,
    "returntype": 0x24,
    "security": 0x1C,
    "GetSet": 0x30,
}

bounded_functions_offsets = {"name": 0x4, "security": 0x1C, "function": 0x40}


def GetDataModel() -> int:
    print(">>>>> Starting DataModel scan ! <<<<<")
    guiroot_pattern = (
        b"\\x47\\x75\\x69\\x52\\x6F\\x6F\\x74\\x00\\x47\\x75\\x69\\x49\\x74\\x65\\x6D"
    )
    guiroot_address = roblox.Program.pattern_scan_all(guiroot_pattern)
    if guiroot_address != 0:
        RawDataModel = roblox.DRP(guiroot_address + 0x28)
        DataModel = RawDataModel + 0xC
        return DataModel
    else:
        return 0


def float_to_hex(f):
    return hex(struct.unpack("<I", struct.pack("<f", f))[0])


def write_str(text: str, Address: int):
    current_char = Address
    mylen = len(text)
    if mylen < 16:
        for char in text:
            roblox.Program.write_char(current_char, char)
            current_char = current_char + 1
    else:
        newmemory = pymem.pymem.memory.allocate_memory(
            roblox.Program.process_handle, mylen
        )
        current_char = newmemory
        for char in text:
            roblox.Program.write_char(current_char, char)
            current_char = current_char + 1
        roblox.Program.write_int(Address, newmemory)
    roblox.Program.write_int(Address + 0x10, mylen)


class Call0Arg:
    def __init__(self, returnType=""):
        self.addr = 0
        self.returnType = returnType

    def allocate(self):
        self.addr = roblox.Program.allocate(100)
        self.InstanceAddress = 0
        self.FunctionAddress = 0
        HexArray = ""
        MovIntoEcxOp = "B9" + roblox.hex2le(roblox.d2h(self.InstanceAddress))
        CallOp = "E8" + roblox.hex2le(
            roblox.calcjmpop(
                roblox.d2h(self.FunctionAddress), roblox.d2h(self.addr + 5)
            )
        )
        if self.returnType == "float":
            StoreOp = "D9 1D" + roblox.hex2le(roblox.d2h(self.addr + 0x30))
        else:
            StoreOp = "A3" + roblox.hex2le(roblox.d2h(self.addr + 0x30))
        RetOp = "C3"
        HexArray = MovIntoEcxOp + CallOp + StoreOp + RetOp
        roblox.Program.write_bytes(
            self.addr, bytes.fromhex(HexArray), roblox.gethexc(HexArray)
        )

    def write(self, instanceAddress: int, functionAddress: int):
        self.InstanceAddress = instanceAddress
        self.FunctionAddress = functionAddress
        MovIntoEcxOp = roblox.hex2le(roblox.d2h(self.InstanceAddress))
        CallOp = roblox.hex2le(
            roblox.calcjmpop(
                roblox.d2h(self.FunctionAddress), roblox.d2h(self.addr + 5)
            )
        )
        roblox.Program.write_bytes(self.addr + 0x1, bytes.fromhex(MovIntoEcxOp), 4)
        roblox.Program.write_bytes(self.addr + 0x6, bytes.fromhex(CallOp), 4)

    def call(self):
        roblox.Program.start_thread(self.addr)
        if self.returnType == "float":
            returnValue = roblox.Program.read_float(self.addr + 0x30)
        if self.returnType == "string":
            returnValue = roblox.ReadInstaceString(self.addr + 0x30)
        else:
            returnValue = roblox.DRP(self.addr + 0x30)
        return returnValue

    def free(self):
        roblox.Program.free(self.addr)


getPropertyFuncs = dict()


class NameMap:
    def __init__(self):
        self.addr = roblox.getAddressFromName(instance_offsets["name_map"])
        self.namelist = {}
        while roblox.Program.read_int(self.addr) != 0:
            self.namelist[
                roblox.ReadInstaceString(self.addr)
            ] = roblox.Program.read_int(self.addr)
            self.addr += 4

    """
    Returns the Address of a string from the NameMap
    """

    def get(self, text: str) -> int:
        if text in self.namelist.keys():
            return self.namelist[text]
        else:
            return 0


nameMap = NameMap()


def SetupOptimizations():
    getFloat = Call0Arg("float")
    getNormal = Call0Arg("")
    getString = Call0Arg("string")
    getVector3 = Call0Arg("Vector3")
    getInt64 = Call0Arg("int64")
    getObject = Call0Arg("Object")
    getPropertyFuncs[""] = getNormal
    getPropertyFuncs["float"] = getFloat
    getPropertyFuncs["string"] = getString
    getPropertyFuncs["Vector3"] = getVector3
    getPropertyFuncs["int64"] = getInt64
    getPropertyFuncs["Object"] = getObject
    for func in getPropertyFuncs:
        getPropertyFuncs[func].allocate()


def FreeOptimizations():
    for func in getPropertyFuncs:
        getPropertyFuncs[func].free()


class GetSetImpl:
    def __init__(self, Address=0) -> None:
        self.addr = Address

    def GetAddress(self) -> int:
        return self.addr

    def Get(self) -> int:
        return roblox.DRP(self.addr + 0x8)

    def Set(self) -> int:
        return roblox.DRP(self.addr + 0x18)


class PropertyDescriptor:
    def __init__(self, Address) -> None:
        self.addr = Address

    def GetAddress(self) -> int:
        return self.addr

    def GetName(self) -> str:
        addr = self.GetAddress()
        return roblox.ReadNormalString(
            roblox.DRP(addr + property_descriptor_offsets["name"])
        )

    def GetReturnValue(self) -> str:
        return roblox.ReadInstaceString(
            roblox.DRP(self.GetAddress() + property_descriptor_offsets["returnType"])
            + 0x4
        )

    def GetSecurity(self) -> int:
        return roblox.Program.read_int(
            self.addr + property_descriptor_offsets["security"]
        )

    def SetSecurity(self, new_security) -> None:
        roblox.Program.write_int(
            self.addr + property_descriptor_offsets["security"], new_security
        )

    def GetSet(self):
        return GetSetImpl(roblox.DRP(self.addr + property_descriptor_offsets["GetSet"]))


rbx_sec_enum = {
    -10: None,
    -1: None,
    0: None,
    1: "Plugin",
    2: "None",
    3: "RobloxPlace",
    4: "RobloxScript",
    5: "LocalUser",
    6: "Roblox",
    7: "Roblox (All permission)",
    8: "WritePlayer",
}


class BoundedFunc:
    def __init__(self, Address=0) -> None:
        self.addr = Address

    def GetAddress(self) -> int:
        return self.addr

    def GetName(self) -> str:
        addr = self.GetAddress()
        return roblox.ReadNormalString(
            roblox.DRP(addr + bounded_functions_offsets["name"])
        )

    def GetFunc(self) -> int:
        return roblox.DRP(self.addr + bounded_functions_offsets["function"])

    def GetSecurity(self) -> int:
        return roblox.Program.read_int(
            self.addr + bounded_functions_offsets["security"]
        )

    def SetSecurity(self, new_security):
        if type(new_security) == int:
            roblox.Program.write_int(
                self.addr + bounded_functions_offsets["security"], new_security
            )
        else:
            print("Correct syntax: SetSecurity(int)")

    def GetSecurityName(self) -> str:
        return rbx_sec_enum[
            roblox.Program.read_int(self.addr + bounded_functions_offsets["security"])
        ]


class BoundedYieldFunc(BoundedFunc):
    def __init__(self, Address=0) -> None:
        super().__init__(Address)

    def GetFunc(self) -> int:
        return roblox.DRP(self.addr + 0x38)


class EventDesc:
    def __init__(self, address=0) -> None:
        self.addr = address

    def GetAddress(self) -> int:
        return self.addr

    def GetName(self) -> str:
        addr = self.GetAddress()
        return roblox.ReadNormalString(roblox.DRP(addr + 0x4))


shared_prop = [
    "Archivable",
    "Attributes",
    "AttributesReplicate",
    "AttributesSerialize",
    "ClassName",
    "DataCost",
    "HistoryId",
    "Name",
    "Parent",
    "PropertyStatusStudio",
    "RobloxLocked",
    "Tags",
    "UniqueId",
    "Version",
    "archivable",
    "className",
    "numExpectedDirectChildren",
    "SourceAssetId",
]

supported_returntypes = [
    "float",
    "int",
    "double",
    "string",
    "bool",
    "Vector3",
]

disabled_props = [
    "AttributesReplicate",
    "AttributesSerialize",
    "CollisionGroups",
    "DataCost",
]


class Instance:
    def __init__(self, otherType=0) -> None:
        if type(otherType) == int:
            self.addr = otherType
        elif type(otherType) == Instance:
            self.addr = otherType.getAddress()

    def __bool__(self) -> bool:
        if self.addr > 0:
            return True
        return False

    def __equal__(self, other) -> bool:
        if type(other) == int:
            return self.addr == int
        if type(other) == Instance:
            return self.addr == other.addr

    def getAddress(self):
        return self.addr

    def GetName(self) -> str:
        addr = self.getAddress()
        return roblox.ReadInstaceString(addr + instance_offsets["name"])

    def HasChildren(self) -> bool:
        child_list = roblox.DRP(self.addr + instance_offsets["childlist"])
        if child_list == 0:
            return False
        else:
            return True

    def GetChildren(self) -> list[Instance]:
        children = []
        if self.addr != 0:
            child_list = roblox.DRP(self.addr + instance_offsets["childlist"])
            if child_list != 0:
                child_begin = roblox.DRP(child_list)
                end_child = roblox.DRP(child_list + 0x4)
                while child_begin != end_child:
                    current_instance = roblox.DRP(child_begin)
                    if current_instance != 0:
                        children.append(Instance(current_instance))
                        child_begin = child_begin + 8
        return children

    def GetParent(self) -> Instance:
        if roblox.DRP(self.addr + instance_offsets["parent"]) != 0:
            return Instance(roblox.DRP(self.addr + instance_offsets["parent"]))
        return Instance(0)

    def FindFirstChild(self, name) -> Instance:
        for child in self.GetChildren():
            if child.GetName() == name:
                return child
        return Instance(0)

    def isA(self, className: str) -> bool:
        if self.GetClassName() == className:
            return True
        return False

    def FindFirstChildOfClass(self, className: str) -> Instance:
        for child in self.GetChildren():
            if child.isA(className):
                return child
        return Instance(0)

    def GetClassDescriptor(self) -> list[Instance]:
        classDescriptor = roblox.DRP(self.addr + instance_offsets["class_descriptor"])
        return classDescriptor

    def GetPropertyDescriptors(self) -> list[PropertyDescriptor]:
        prop_begin = roblox.DRP(
            self.GetClassDescriptor() + instance_offsets["property_descriptor"]
        )
        prop_end = roblox.DRP(
            self.GetClassDescriptor() + instance_offsets["property_descriptor"] + 0x4
        )
        children = []
        if prop_begin == 0 or prop_end == 0:
            return []
        while prop_begin != prop_end:
            current_prop = roblox.DRP(prop_begin)
            if current_prop != 0:
                children.append(PropertyDescriptor(current_prop))
            prop_begin += 4
        return children

    def GetEventDescs(self) -> list[EventDesc]:
        prop_begin = roblox.DRP(
            self.GetClassDescriptor() + instance_offsets["event_descriptor"]
        )
        prop_end = roblox.DRP(
            self.GetClassDescriptor() + instance_offsets["event_descriptor"] + 0x4
        )
        children = []
        if prop_begin == 0 or prop_end == 0:
            return []
        while prop_begin != prop_end:
            current_prop = roblox.DRP(prop_begin)
            if current_prop != 0:
                children.append(EventDesc(current_prop))
            prop_begin += 4
        return children

    def GetEventDesc(self, name: str) -> PropertyDescriptor:
        for prop in self.GetEventDescs():
            if prop.GetName() == name:
                return prop
        return 0

    def GetPropertyDescriptor(self, name: str) -> PropertyDescriptor:
        for prop in self.GetPropertyDescriptors():
            if prop.GetName() == name:
                return prop
        return 0

    def GetBoundedFuncs(self) -> list[BoundedFunc]:
        prop_begin = roblox.DRP(
            self.GetClassDescriptor() + instance_offsets["boundedfunctions"]
        )
        prop_end = roblox.DRP(
            self.GetClassDescriptor() + instance_offsets["boundedfunctions"] + 0x4
        )
        children = []
        if prop_begin == 0 or prop_end == 0:
            return []
        while prop_begin != prop_end:
            current_prop = roblox.DRP(prop_begin)
            if current_prop != 0:
                children.append(BoundedFunc(current_prop))
            prop_begin += 4
        return children

    def GetBoundFunction(self, name: str) -> BoundedFunc:
        for func in self.GetBoundedFuncs():
            if func.GetName() == name:
                return func
        return 0

    def GetClassName(self) -> str:
        return roblox.ReadInstaceString(self.GetClassDescriptor() + 0x4)

    def GetProperty(self, name):
        propertyDescriptor = self.GetPropertyDescriptor(name)
        ReturnType = roblox.ReadInstaceString(
            roblox.DRP(
                propertyDescriptor.GetAddress()
                + property_descriptor_offsets["returntype"]
            )
            + 0x4
        )
        if ReturnType in getPropertyFuncs:
            getfunc = getPropertyFuncs[ReturnType]
            if name == "Character":
                FunctionAddress = roblox.DRP(
                    roblox.DRP(self.GetPropertyDescriptor(name).GetAddress() + 0x34)
                    + 0x8
                )
                getfunc.write(self.addr, FunctionAddress)
            else:
                getfunc.write(self.addr, propertyDescriptor.GetSet().Get())
            return getfunc.call()
        else:
            return ReturnType + " Not Implemented"

    def SetProperty(self, name, arg):
        NewMemoryRegion = roblox.Program.allocate(100)
        NewMemAddress = NewMemoryRegion
        InstanceAddress = self.addr
        if name == "Parent":
            FunctionAddress = roblox.DRP(
                roblox.DRP(self.GetPropertyDescriptor(name).GetAddress() + 0x34) + 0x10
            )
        else:
            FunctionAddress = self.GetPropertyDescriptor(name).GetSet().Set()
        HexArray = ""
        MovIntoEcxOp = "B9" + roblox.hex2le(roblox.d2h(InstanceAddress))
        PushOP = "68" + roblox.hex2le(roblox.d2h(arg))
        CallOp = "E8" + roblox.hex2le(
            roblox.calcjmpop(
                roblox.d2h(FunctionAddress), roblox.d2h(NewMemAddress + 10)
            )
        )
        StoreOp = "A3" + roblox.hex2le(roblox.d2h(NewMemAddress + 0x40))
        RetOp = "C3"
        HexArray = MovIntoEcxOp + PushOP + CallOp + StoreOp + RetOp
        roblox.Program.write_bytes(
            NewMemAddress, bytes.fromhex(HexArray), roblox.gethexc(HexArray)
        )
        roblox.Program.start_thread(NewMemAddress)

    def GetDescendants(self) -> list[Instance]:
        descendants = []
        for child in self.GetChildren():
            descendants.append(child)
            descendants += child.GetDescendants()
        return descendants

    def Destroy(self):
        NewMemoryRegion = roblox.Program.allocate(100)
        NewMemAddress = NewMemoryRegion
        InstanceAddress = self.getAddress()
        FunctionAddress = self.GetBoundFunction("Destroy").GetFunc()
        HexArray = ""
        MovIntoEcxOp = "B9" + roblox.hex2le(roblox.d2h(InstanceAddress))
        CallOp = "E8" + roblox.hex2le(
            roblox.calcjmpop(roblox.d2h(FunctionAddress), roblox.d2h(NewMemAddress + 5))
        )
        StoreOp = "A3" + roblox.hex2le(roblox.d2h(NewMemAddress + 0x30))
        RetOp = "C3"
        HexArray = MovIntoEcxOp + CallOp + StoreOp + RetOp
        roblox.Program.write_bytes(
            NewMemAddress, bytes.fromhex(HexArray), roblox.gethexc(HexArray)
        )
        roblox.Program.start_thread(NewMemAddress)
        roblox.Program.free(NewMemAddress)

    def Clone(self):
        NewMemoryRegion = roblox.Program.allocate(100)
        NewMemAddress = NewMemoryRegion
        InstanceAddress = self.getAddress()
        FunctionAddress = self.GetBoundFunction("Clone").GetFunc()
        HexArray = ""
        MovIntoEcxOp = "B9" + roblox.hex2le(roblox.d2h(InstanceAddress))
        PushEcx = "FF 71 08"
        CallOp = "E8" + roblox.hex2le(
            roblox.calcjmpop(roblox.d2h(FunctionAddress), roblox.d2h(NewMemAddress + 8))
        )
        StoreOp = "A3" + roblox.hex2le(roblox.d2h(NewMemAddress + 0x30))
        RetOp = "C3"
        HexArray = MovIntoEcxOp + PushEcx + CallOp + StoreOp + RetOp
        roblox.Program.write_bytes(
            NewMemAddress, bytes.fromhex(HexArray), roblox.gethexc(HexArray)
        )
        roblox.Program.start_thread(NewMemAddress)
        returnValue = roblox.DRP(roblox.DRP(NewMemAddress + 0x30))
        roblox.Program.free(NewMemAddress)
        return returnValue

    def new(self, className: str) -> Instance:
        FunctionToCall = roblox.getAddressFromName(instance_offsets["new"])
        NewMemoryRegion = roblox.Program.allocate(100)
        returnStruct = roblox.Program.allocate(4)
        NewMemAddress = NewMemoryRegion
        HexArray = ""
        MovIntoEcxOp = "B9" + roblox.hex2le(roblox.d2h(returnStruct))
        MovIntoEdxOp = "BA" + roblox.hex2le(roblox.d2h(nameMap.get(className)))
        PushOP = "6A 03"
        CallOp = "E8" + roblox.hex2le(
            roblox.calcjmpop(roblox.d2h(FunctionToCall), roblox.d2h(NewMemAddress + 12))
        )
        AddOp = "83 C4 04"
        RetOp = "C3"
        HexArray = PushOP + MovIntoEcxOp + MovIntoEdxOp + CallOp + AddOp + RetOp
        roblox.Program.write_bytes(
            NewMemAddress, bytes.fromhex(HexArray), roblox.gethexc(HexArray)
        )
        roblox.Program.start_thread(NewMemAddress)
        retInstance = Instance(roblox.DRP(returnStruct))
        roblox.Program.free(NewMemAddress)
        roblox.Program.free(returnStruct)
        return retInstance


shared_instances = {"Game": 0, "Workspace": 0, "Players": 0}


class Vector3:
    def __init__(self, x=0.0, y=0.0, z=0.0) -> None:
        self.x = x
        self.y = y
        self.z = z

    def fromMemory(self, address) -> Vector3:
        self.x = roblox.Program.read_float(address)
        self.y = roblox.Program.read_float(address + 0x4)
        self.z = roblox.Program.read_float(address + 0x8)
        return self

    def __repr__(self) -> str:
        return (
            "Vector 3 with values : "
            + str(self.x)
            + " "
            + str(self.y)
            + " "
            + str(self.z)
        )


class Camera(Instance):
    def __init__(self, otherType) -> None:
        if type(otherType) == int:
            super().__init__(otherType)
        elif type(otherType) == Instance:
            super().__init__(otherType.getAddress())

    def allocate_w2s(self):
        self.w2s_addr = roblox.Program.allocate(100)
        self.returnStruct = roblox.Program.allocate(4)
        InstanceAddress = self.getAddress()
        FunctionAddress = self.GetBoundFunction("WorldToScreenPoint").GetFunc()
        HexArray = ""
        MovIntoEcxOp = "B9" + roblox.hex2le(roblox.d2h(InstanceAddress))
        PushOPX = "68" + roblox.hex2le(roblox.d2h(int(float_to_hex(0), 16)))
        PushOPY = "68" + roblox.hex2le(roblox.d2h(int(float_to_hex(0), 16)))
        PushOPZ = "68" + roblox.hex2le(roblox.d2h(int(float_to_hex(0), 16)))
        PushOP2 = "68" + roblox.hex2le(roblox.d2h(self.w2s_addr + 0x30))
        CallOp = "E8" + roblox.hex2le(
            roblox.calcjmpop(
                roblox.d2h(FunctionAddress), roblox.d2h(self.w2s_addr + 25)
            )
        )
        RetOp = "C3"
        HexArray = MovIntoEcxOp + PushOPX + PushOPY + PushOPZ + PushOP2 + CallOp + RetOp
        roblox.Program.write_bytes(
            self.w2s_addr, bytes.fromhex(HexArray), roblox.gethexc(HexArray)
        )

    def write_w2s(self, x, y, z):
        x_value = roblox.hex2le(roblox.d2h(int(float_to_hex(x), 16)))
        y_value = roblox.hex2le(roblox.d2h(int(float_to_hex(y), 16)))
        z_value = roblox.hex2le(roblox.d2h(int(float_to_hex(z), 16)))
        roblox.Program.write_bytes(self.w2s_addr + 6, bytes.fromhex(x_value), 4)
        roblox.Program.write_bytes(self.w2s_addr + 11, bytes.fromhex(y_value), 4)
        roblox.Program.write_bytes(self.w2s_addr + 16, bytes.fromhex(z_value), 4)

    def call_w2s(self):
        roblox.Program.start_thread(self.w2s_addr)
        return Vector3().fromMemory(roblox.DRP(roblox.DRP(self.w2s_addr + 0x30)) + 0x8)

    def free_w2s(self):
        roblox.Program.free(self.w2s_addr)


class Highlight(Instance):
    def __init__(self, otherType=0) -> None:
        super().__init__(otherType)

    def SetFillColor(self, newColor: Vector3):
        NewMemoryRegion = roblox.Program.allocate(100)
        NewMemAddress = NewMemoryRegion
        struct = roblox.Program.allocate(12)
        InstanceAddress = self.getAddress()
        FunctionAddress = self.GetPropertyDescriptor("FillColor").GetSet().Set()
        roblox.Program.write_float(struct, newColor.x)
        roblox.Program.write_float(struct + 4, newColor.y)
        roblox.Program.write_float(struct + 8, newColor.z)
        HexArray = ""
        MovIntoEcxOp = "B9" + roblox.hex2le(roblox.d2h(InstanceAddress))
        PushOPX = "68" + roblox.hex2le(roblox.d2h(struct))
        CallOp = "E8" + roblox.hex2le(
            roblox.calcjmpop(
                roblox.d2h(FunctionAddress), roblox.d2h(NewMemAddress + 10)
            )
        )
        StoreOp = "A3" + roblox.hex2le(roblox.d2h(NewMemAddress + 0x30))
        RetOp = "C3"
        HexArray = MovIntoEcxOp + PushOPX + CallOp + StoreOp + RetOp
        roblox.Program.write_bytes(
            NewMemAddress, bytes.fromhex(HexArray), roblox.gethexc(HexArray)
        )
        roblox.Program.start_thread(NewMemAddress)
        returnValue = roblox.DRP(NewMemAddress + 0x30)
        roblox.Program.free(NewMemAddress)
        return returnValue


class Player(Instance):
    def __init__(self, otherType=0) -> None:
        super().__init__(otherType)

    def GetCharacter(self) -> Instance:
        return shared_instances["Workspace"].FindFirstChild(self.GetName())


class Players(Instance):
    def __init__(self, instance: Instance) -> None:
        super().__init__(instance.getAddress())

    def GetLocalPlayer(self) -> Instance:
        return self.GetChildren()[0]

    def GetLocalPlayerChar(self, workspace) -> Instance:
        if type(workspace) != Instance:
            print("nah")
        else:
            LP = self.GetLocalPlayer()
            return workspace.FindFirstChild(LP.GetName())

    def GetAllPlayers(self) -> list[Player]:
        plrs = []
        for players in self.GetChildren():
            plrs.append(players)
        return plrs


DataModel = Instance(GetDataModel())

print(roblox.d2h(DataModel.getAddress()))
Players = Players(DataModel.FindFirstChildOfClass("Players"))
workspace = DataModel.GetChildren()[0]

localPlayer = Players.GetLocalPlayer()
playerChar = workspace.FindFirstChild(localPlayer.GetName())


def BreakJoints(character: Instance):
    NewMemoryRegion = roblox.Program.allocate(100)
    NewMemAddress = NewMemoryRegion
    InstanceAddress = character.getAddress()
    FunctionAddress = character.GetBoundFunction("BreakJoints").GetFunc()
    HexArray = ""
    MovIntoEcxOp = "B9" + roblox.hex2le(roblox.d2h(InstanceAddress))
    CallOp = "E8" + roblox.hex2le(
        roblox.calcjmpop(roblox.d2h(FunctionAddress), roblox.d2h(NewMemAddress + 5))
    )
    StoreOp = "A3" + roblox.hex2le(roblox.d2h(NewMemAddress + 0x30))
    RetOp = "C3"
    HexArray = MovIntoEcxOp + CallOp + StoreOp + RetOp
    roblox.Program.write_bytes(
        NewMemAddress, bytes.fromhex(HexArray), roblox.gethexc(HexArray)
    )
    roblox.Program.start_thread(NewMemAddress)
    roblox.Program.free(NewMemAddress)


def MoveTo(character: Instance):
    NewMemoryRegion = roblox.Program.allocate(100)
    NewMemAddress = NewMemoryRegion
    InstanceAddress = character.getAddress()
    FunctionAddress = character.GetBoundFunction("MoveTo").GetFunc()
    testx = 3.0
    testy = 4.0
    testz = 5.0
    HexArray = ""
    MovIntoEcxOp = "B9" + roblox.hex2le(roblox.d2h(InstanceAddress))
    PushOPX = "68" + roblox.hex2le(roblox.d2h(int(float_to_hex(testx), 16)))
    PushOPY = "68" + roblox.hex2le(roblox.d2h(int(float_to_hex(testy), 16)))
    PushOPZ = "68" + roblox.hex2le(roblox.d2h(int(float_to_hex(testz), 16)))
    CallOp = "E8" + roblox.hex2le(
        roblox.calcjmpop(roblox.d2h(FunctionAddress), roblox.d2h(NewMemAddress + 20))
    )
    StoreOp = "A3" + roblox.hex2le(roblox.d2h(NewMemAddress + 0x30))
    RetOp = "C3"
    HexArray = MovIntoEcxOp + PushOPZ + PushOPY + PushOPX + CallOp + StoreOp + RetOp
    roblox.Program.write_bytes(
        NewMemAddress, bytes.fromhex(HexArray), roblox.gethexc(HexArray)
    )
    print(roblox.d2h(NewMemAddress))
    roblox.Program.start_thread(NewMemAddress)
    returnValue = roblox.DRP(NewMemAddress + 0x30)
    roblox.Program.free(NewMemAddress)
    return returnValue


MoveTo(playerChar)
